@@ -112,7 +112,7 @@ def canCoinstall(arch1, arch2):
 def archDifference(myarch, targetarch):
     if myarch == targetarch:
         return 1
-    if arches.has_key(myarch):
+    if myarch in arches:
         ret = archDifference(arches[myarch], targetarch)
         if ret != 0:
             return ret + 1
@@ -127,13 +127,13 @@ def isMultiLibArch(arch=None):
     if arch is None:
         arch = canonArch
 
-    if not arches.has_key(arch): # or we could check if it is noarch
+    if arch not in arches: # or we could check if it is noarch
         return 0
     
-    if multilibArches.has_key(arch):
+    if arch in multilibArches:
         return 1
         
-    if multilibArches.has_key(arches[arch]):
+    if arches[arch] in multilibArches:
         return 1
     
     return 0
@@ -188,7 +188,7 @@ def getArchList(thisarch=None):
         thisarch = canonArch
     
     archlist = [thisarch]
-    while arches.has_key(thisarch):
+    while thisarch in arches:
         thisarch = arches[thisarch]
         archlist.append(thisarch)
 
@@ -202,15 +202,19 @@ def getArchList(thisarch=None):
         archlist.append('noarch')
     return archlist
     
-        
+def _try_read_cpuinfo():
+    """ Try to read /proc/cpuinfo ... if we can't ignore errors (ie. proc not
+        mounted). """
+    try:
+        lines = open("/proc/cpuinfo", "r").readlines()
+        return lines
+    except:
+        return []
 
 def getCanonX86Arch(arch):
     # 
     if arch == "i586":
-        f = open("/proc/cpuinfo", "r")
-        lines = f.readlines()
-        f.close()
-        for line in lines:
+        for line in _try_read_cpuinfo():
             if line.startswith("model name") and line.find("Geode(TM)") != -1:
                 return "geode"
         return arch
@@ -219,10 +223,7 @@ def getCanonX86Arch(arch):
         return arch
 
     # if we're i686 and AuthenticAMD, then we should be an athlon
-    f = open("/proc/cpuinfo", "r")
-    lines = f.readlines()
-    f.close()
-    for line in lines:
+    for line in _try_read_cpuinfo():
         if line.startswith("vendor") and line.find("AuthenticAMD") != -1:
             return "athlon"
         # i686 doesn't guarantee cmov, but we depend on it
@@ -237,10 +238,7 @@ def getCanonPPCArch(arch):
         return arch
 
     machine = None
-    f = open("/proc/cpuinfo", "r")
-    lines = f.readlines()
-    f.close()
-    for line in lines:
+    for line in _try_read_cpuinfo():
         if line.find("machine") != -1:
             machine = line.split(':')[1]
             break
@@ -256,10 +254,7 @@ def getCanonPPCArch(arch):
 def getCanonSPARCArch(arch):
     # Deal with sun4v, sun4u, sun4m cases
     SPARCtype = None
-    f = open("/proc/cpuinfo", "r")
-    lines = f.readlines()
-    f.close()
-    for line in lines:
+    for line in _try_read_cpuinfo():
         if line.startswith("type"):
             SPARCtype = line.split(':')[1]
             break
@@ -285,10 +280,7 @@ def getCanonX86_64Arch(arch):
         return arch
 
     vendor = None
-    f = open("/proc/cpuinfo", "r")
-    lines = f.readlines()
-    f.close()
-    for line in lines:
+    for line in _try_read_cpuinfo():
         if line.startswith("vendor_id"):
             vendor = line.split(':')[1]
             break
@@ -330,9 +322,9 @@ canonArch = getCanonArch()
 
 # this gets you the "compat" arch of a biarch pair
 def getMultiArchInfo(arch = canonArch):
-    if multilibArches.has_key(arch):
+    if arch in multilibArches:
         return multilibArches[arch]
-    if arches.has_key(arch) and arches[arch] != "noarch":
+    if arch in arches and arches[arch] != "noarch":
         return getMultiArchInfo(arch = arches[arch])
     return None
 
@@ -361,7 +353,7 @@ def getBaseArch(myarch=None):
     if not myarch:
         myarch = canonArch
 
-    if not arches.has_key(myarch): # this is dumb, but <shrug>
+    if myarch not in arches: # this is dumb, but <shrug>
         return myarch
 
     if myarch.startswith("sparc64"):
@@ -370,12 +362,12 @@ def getBaseArch(myarch=None):
         return "ppc"
         
     if isMultiLibArch(arch=myarch):
-        if multilibArches.has_key(myarch):
+        if myarch in multilibArches:
             return myarch
         else:
             return arches[myarch]
     
-    if arches.has_key(myarch):
+    if myarch in arches:
         basearch = myarch
         value = arches[basearch]
         while value != 'noarch':
@@ -397,7 +389,7 @@ class ArchStorage(object):
         self.multilib = False
         self.setup_arch()
 
-    def setup_arch(self, arch=None):
+    def setup_arch(self, arch=None, archlist_includes_compat_arch=True):
         if arch:
             self.canonarch = arch
         else:
@@ -405,6 +397,14 @@ class ArchStorage(object):
         
         self.basearch = getBaseArch(myarch=self.canonarch)
         self.archlist = getArchList(thisarch=self.canonarch)
+        
+        if not archlist_includes_compat_arch: # - do we bother including i686 and below on x86_64
+            limit_archlist = []
+            for a in self.archlist:
+                if isMultiLibArch(a) or a == 'noarch':
+                    limit_archlist.append(a)
+            self.archlist = limit_archlist
+            
         self.bestarch = getBestArch(myarch=self.canonarch)
         self.compatarches = getMultiArchInfo(arch=self.canonarch)
         self.multilib = isMultiLibArch(arch=self.canonarch)
