@@ -39,6 +39,8 @@ if not _use_iniparse:
     from ConfigParser import ConfigParser
 import rpmUtils.transaction
 import Errors
+import types
+from misc import get_uuid, read_in_items_from_dot_dir
 
 # Alter/patch these to change the default checking...
 __pkgs_gpgcheck_default__ = False
@@ -53,9 +55,11 @@ class Option(object):
     definition easy and consise.
     '''
 
-    def __init__(self, default=None):
+    def __init__(self, default=None, parse_default=False):
         self._setattrname()
         self.inherit = False
+        if parse_default:
+            default = self.parse(default)
         self.default = default
 
     def _setattrname(self):
@@ -85,14 +89,14 @@ class Option(object):
         @param value: The value to set the option to.
         @return: Nothing.
         '''
-        # Only try to parse if its a string
+        # Only try to parse if it's a string
         if isinstance(value, basestring):
             try:
                 value = self.parse(value)
             except ValueError, e:
                 # Add the field name onto the error
-                raise ValueError('Error parsing %r: %s' % (value, str(e)))
-
+                raise ValueError('Error parsing "%s = %r": %s' % (self._optname,
+                                                                 value, str(e)))
         setattr(obj, self._attrname, value)
 
     def setup(self, obj, name):
@@ -102,6 +106,7 @@ class Option(object):
         @param obj: BaseConfig (or subclass) instance.
         @param name: Name of the option.
         '''
+        self._optname = name
         setattr(obj, self._attrname, copy.copy(self.default))
 
     def clone(self):
@@ -152,13 +157,14 @@ class ListOption(Option):
     An option containing a list of strings.
     """
 
-    def __init__(self, default=None):
+    def __init__(self, default=None, parse_default=False):
         if default is None:
             default = []
-        super(ListOption, self).__init__(default)
+        super(ListOption, self).__init__(default, parse_default)
 
     def parse(self, s):
-        """Converts a string from the config file to a workable list
+        """Converts a string from the config file to a workable list, parses
+           globdir: paths as foo.d-style dirs
 
         Commas and spaces are used as separators for the list
         """
@@ -166,7 +172,15 @@ class ListOption(Option):
         # to sub the \n with a space and then read the lines
         s = s.replace('\n', ' ')
         s = s.replace(',', ' ')
-        return s.split()
+        results = []
+        for item in s.split():
+            if item.startswith('glob:'):
+                thisglob = item.replace('glob:', '')
+                results.extend(read_in_items_from_dot_dir(thisglob))
+                continue
+            results.append(item)
+
+        return results
 
     def tostring(self, value):
         return '\n '.join(value)
@@ -214,8 +228,9 @@ class UrlListOption(ListOption):
     Option for handling lists of URLs with validation of the URL scheme.
     '''
 
-    def __init__(self, default=None, schemes=('http', 'ftp', 'file', 'https')):
-        super(UrlListOption, self).__init__(default)
+    def __init__(self, default=None, schemes=('http', 'ftp', 'file', 'https'),
+                 parse_default=False):
+        super(UrlListOption, self).__init__(default, parse_default)
 
         # Hold a UrlOption instance to assist with parsing
         self._urloption = UrlOption(schemes=schemes)
@@ -225,8 +240,15 @@ class UrlListOption(ListOption):
         s = s.replace('\n', ' ')
         s = s.replace(',', ' ')
         items = [ item.replace(' ', '%20') for item in shlex.split(s) ]
-        s = ' '.join(items)
-        for url in super(UrlListOption, self).parse(s):
+        tmp = []
+        for item in items:
+            if item.startswith('glob:'):
+                thisglob = item.replace('glob:', '')
+                tmp.extend(read_in_items_from_dot_dir(thisglob))
+                continue
+            tmp.append(item)
+
+        for url in super(UrlListOption, self).parse(' '.join(tmp)):
             out.append(self._urloption.parse(url))
         return out
 
@@ -598,8 +620,10 @@ class StartupConf(BaseConfig):
     pluginconfpath = ListOption(['/etc/yum/pluginconf.d'])
     gaftonmode = BoolOption(False)
     syslog_ident = Option()
-    syslog_facility = Option('LOG_DAEMON')
-
+    syslog_facility = Option('LOG_USER')
+    syslog_device = Option('/dev/log')
+    persistdir = Option('/var/lib/yum')
+    
 class YumConf(StartupConf):
     '''
     Configuration option definitions for yum.conf\'s [main] section.
@@ -610,7 +634,7 @@ class YumConf(StartupConf):
     recent = IntOption(7, range_min=0)
 
     cachedir = Option('/var/cache/yum')
-    persistdir = Option('/var/lib/yum')
+
     keepcache = BoolOption(True)
     logfile = Option('/var/log/yum.log')
     reposdir = ListOption(['/etc/yum/repos.d', '/etc/yum.repos.d'])
@@ -652,6 +676,7 @@ class YumConf(StartupConf):
     enabled = BoolOption(True)
     enablegroups = BoolOption(True)
     enable_group_conditionals = BoolOption(True)
+    groupremove_leaf_only = BoolOption(False)
     group_package_types = ListOption(['mandatory', 'default'])
     
     timeout = FloatOption(30.0) # FIXME: Should use variation of SecondsOption
@@ -706,8 +731,36 @@ class YumConf(StartupConf):
     history_record = BoolOption(True)
     history_record_packages = ListOption(['yum', 'rpm'])
 
+    rpmverbosity = Option('info')
+
+    protected_packages = ListOption("yum, glob:/etc/yum/protected.d/*.conf",
+                                    parse_default=True)
+    exit_on_lock = BoolOption(False)
+    
     _reposlist = []
 
+    def dump(self):
+        output = '[main]\n'
+        # we exclude all vars which start with _ or are in this list:
+        excluded_vars = ('cfg', 'uid', 'yumvar', 'progress_obj', 'failure_obj',
+                         'disable_excludes', 'config_file_age', 'config_file_path',
+                         )
+        for attr in dir(self):
+            if attr.startswith('_'):
+                continue
+            if attr in excluded_vars:
+                continue
+            if isinstance(getattr(self, attr), types.MethodType):
+                continue
+            res = getattr(self, attr)
+            if not res and type(res) not in (type(False), type(0)):
+                res = ''
+            if type(res) == types.ListType:
+                res = ',\n   '.join(res)
+            output = output + '%s = %s\n' % (attr, res)
+
+        return output
+
 class RepoConf(BaseConfig):
     '''
     Option definitions for repository INI file sections.
@@ -764,6 +817,7 @@ class RepoConf(BaseConfig):
     sslclientcert = Inherit(YumConf.sslclientcert)
     sslclientkey = Inherit(YumConf.sslclientkey)
 
+    skip_if_unavailable = BoolOption(False)
     
 class VersionGroupConf(BaseConfig):
     pkglist = ListOption()
@@ -805,6 +859,8 @@ def readStartupConfig(configfile, root):
     startupconf._parser = parser
     # setup the release ver here
     startupconf.releasever = _getsysver(startupconf.installroot, startupconf.distroverpkg)
+    uuidfile = '%s/%s/uuid' % (startupconf.installroot, startupconf.persistdir)
+    startupconf.uuid = get_uuid(uuidfile)
 
     return startupconf
 
@@ -823,13 +879,16 @@ def readMainConfig(startupconf):
     yumvars['basearch'] = startupconf.basearch
     yumvars['arch'] = startupconf.arch
     yumvars['releasever'] = startupconf.releasever
-
+    yumvars['uuid'] = startupconf.uuid
+    # Note: We don't setup the FS yumvars here, because we want to be able to
+    #       use the core yumvars in persistdir. Which is the base of FS yumvars.
+    
     # Read [main] section
     yumconf = YumConf()
     yumconf.populate(startupconf._parser, 'main')
 
     # Apply the installroot to directory options
-    for option in ('cachedir', 'logfile', 'persistdir'):
+    def _apply_installroot(yumconf, option):
         path = getattr(yumconf, option)
         ir_path = yumconf.installroot + path
         ir_path = ir_path.replace('//', '/') # os.path.normpath won't fix this and
@@ -837,6 +896,27 @@ def readMainConfig(startupconf):
         ir_path = varReplace(ir_path, yumvars)
         setattr(yumconf, option, ir_path)
     
+    # Read the FS yumvars
+    try:
+        dir_fsvars = yumconf.installroot + "/etc/yum/vars/"
+        fsvars = os.listdir(dir_fsvars)
+    except OSError:
+        fsvars = []
+    for fsvar in fsvars:
+        if os.path.islink(dir_fsvars + fsvar):
+            continue
+        try:
+            val = open(dir_fsvars + fsvar).readline()
+            if val and val[-1] == '\n':
+                val = val[:-1]
+        except (OSError, IOError):
+            continue
+        yumvars[fsvar] = val
+
+    # These can use the above FS yumvars
+    for option in ('cachedir', 'logfile', 'persistdir'):
+        _apply_installroot(yumconf, option)
+
     # Add in some extra attributes which aren't actually configuration values 
     yumconf.yumvar = yumvars
     yumconf.uid = 0
@@ -921,6 +1001,9 @@ def _getsysver(installroot, distroverpkg):
             else:
                 raise Errors.YumBaseError("Error: " + str(e))
         raise Errors.YumBaseError("Error: " + str(e))
+    except rpm.error, e:
+        # This is the "new" code for "cannot open rpmdb", 4.8.0 ish
+        raise Errors.YumBaseError("Error: " + str(e))
     # we're going to take the first one - if there is more than one of these
     # then the user needs a beating
     if idx.count() == 0:
@@ -945,13 +1028,21 @@ def writeRawRepoFile(repo,only=None):
         return
 
     ini = INIConfig(open(repo.repofile))
+    # b/c repoids can have $values in them we need to map both ways to figure
+    # out which one is which
+    section_id = repo.id
+    if repo.id not in ini._sections:
+        for sect in ini._sections.keys():
+            if varReplace(sect, repo.yumvar) == repo.id:
+                section_id = sect
+    
     # Updated the ConfigParser with the changed values    
     cfgOptions = repo.cfg.options(repo.id)
     for name,value in repo.iteritems():
         option = repo.optionobj(name)
         if option.default != value or name in cfgOptions :
             if only == None or name in only:
-                ini[repo.id][name] = option.tostring(value)
+                ini[section_id][name] = option.tostring(value)
     fp =file(repo.repofile,"w")               
     fp.write(str(ini))
     fp.close()
