@@ -22,7 +22,7 @@ import os, os.path
 import glob
 from weakref import proxy as weakref
 
-from sqlutils import sqlite, executeSQL
+from sqlutils import sqlite, executeSQL, sql_esc_glob
 import yum.misc as misc
 import yum.constants
 from yum.constants import *
@@ -56,30 +56,6 @@ _sttxt2stcode = {'Update' : TS_UPDATE,
                  'Obsoleting' : TS_OBSOLETING}
 
 # ---- horrible Copy and paste from sqlitesack ----
-def _sql_esc(pattern):
-    """ Apply SQLite escaping, if needed. Returns pattern and esc. """
-    esc = ''
-    if "_" in pattern or "%" in pattern:
-        esc = ' ESCAPE "!"'
-        pattern = pattern.replace("!", "!!")
-        pattern = pattern.replace("%", "!%")
-        pattern = pattern.replace("_", "!_")
-    return (pattern, esc)
-
-def _sql_esc_glob(patterns):
-    """ Converts patterns to SQL LIKE format, if required (or gives up if
-        not possible). """
-    ret = []
-    for pattern in patterns:
-        if '[' in pattern: # LIKE only has % and _, so [abc] can't be done.
-            return []      # So Load everything
-
-        # Convert to SQL LIKE format
-        (pattern, esc) = _sql_esc(pattern)
-        pattern = pattern.replace("*", "%")
-        pattern = pattern.replace("?", "_")
-        ret.append((pattern, esc))
-    return ret
 
 def _setupHistorySearchSQL(patterns=None, ignore_case=False):
     """Setup need_full and patterns for _yieldSQLDataList, also see if
@@ -104,7 +80,7 @@ def _setupHistorySearchSQL(patterns=None, ignore_case=False):
     if len(patterns) > pat_max:
         patterns = []
     if ignore_case:
-        patterns = _sql_esc_glob(patterns)
+        patterns = sql_esc_glob(patterns)
     else:
         tmp = []
         need_glob = False
@@ -136,6 +112,35 @@ class YumHistoryPackage(PackageObject):
             chk = checksum.split(':')
             self._checksums = [(chk[0], chk[1], 0)] # (type, checksum, id(0,1))
 
+class YumHistoryRpmdbProblem(PackageObject):
+    """ Class representing an rpmdb problem that existed at the time of the
+        transaction. """
+
+    def __init__(self, history, rpid, problem, text):
+        self._history = weakref(history)
+
+        self.rpid = rpid
+        self.problem = problem
+        self.text = text
+
+        self._loaded_P = None
+
+    def __cmp__(self, other):
+        if other is None:
+            return 1
+        ret = cmp(self.problem, other.problem)
+        if ret: return -ret
+        ret = cmp(self.rpid, other.rpid)
+        return ret
+
+    def _getProbPkgs(self):
+        if self._loaded_P is None:
+            self._loaded_P = sorted(self._history._old_prob_pkgs(self.rpid))
+        return self._loaded_P
+
+    packages = property(fget=lambda self: self._getProbPkgs())
+
+
 class YumHistoryTransaction:
     """ Holder for a history transaction. """
 
@@ -152,6 +157,12 @@ class YumHistoryTransaction:
 
         self._loaded_TW = None
         self._loaded_TD = None
+        self._loaded_TS = None
+
+        self._loaded_PROB = None
+
+        self._have_loaded_CMD = False # cmdline can validly be None
+        self._loaded_CMD = None
 
         self._loaded_ER = None
         self._loaded_OT = None
@@ -177,9 +188,29 @@ class YumHistoryTransaction:
         if self._loaded_TD is None:
             self._loaded_TD = sorted(self._history._old_data_pkgs(self.tid))
         return self._loaded_TD
+    def _getTransSkip(self):
+        if self._loaded_TS is None:
+            self._loaded_TS = sorted(self._history._old_skip_pkgs(self.tid))
+        return self._loaded_TS
 
     trans_with = property(fget=lambda self: self._getTransWith())
     trans_data = property(fget=lambda self: self._getTransData())
+    trans_skip = property(fget=lambda self: self._getTransSkip())
+
+    def _getProblems(self):
+        if self._loaded_PROB is None:
+            self._loaded_PROB = sorted(self._history._old_problems(self.tid))
+        return self._loaded_PROB
+
+    rpmdb_problems = property(fget=lambda self: self._getProblems())
+
+    def _getCmdline(self):
+        if not self._have_loaded_CMD:
+            self._have_loaded_CMD = True
+            self._loaded_CMD = self._history._old_cmdline(self.tid)
+        return self._loaded_CMD
+
+    cmdline = property(fget=lambda self: self._getCmdline())
 
     def _getErrors(self):
         if self._loaded_ER is None:
@@ -200,8 +231,12 @@ class YumHistory:
         self._conn = None
         
         self.conf = yum.misc.GenericHolder()
-        self.conf.db_path  = os.path.normpath(root + '/' + db_path)
+        if not os.path.normpath(db_path).startswith(root):
+            self.conf.db_path  = os.path.normpath(root + '/' + db_path)
+        else:
+            self.conf.db_path = os.path.normpath('/' + db_path)
         self.conf.writable = False
+        self.conf.readable = True
 
         if not os.path.exists(self.conf.db_path):
             try:
@@ -227,18 +262,41 @@ class YumHistory:
             except ValueError:
                 continue
 
+            self._db_date = '%s-%s-%s' % (pieces[0], pieces[1], pieces[2])
             self._db_file = d
             break
 
         if self._db_file is None:
             self._create_db_file()
+        
+        # make an addon path for where we're going to stick 
+        # random additional history info - probably from plugins and what-not
+        self.conf.addon_path = self.conf.db_path + '/' + self._db_date
+        if not os.path.exists(self.conf.addon_path):
+            try:
+                os.makedirs(self.conf.addon_path)
+            except (IOError, OSError), e:
+                # some sort of useful thing here? A warning?
+                return
+        else:
+            if os.access(self.conf.addon_path, os.W_OK):
+                self.conf.writable = True
+
 
     def __del__(self):
         self.close()
 
     def _get_cursor(self):
         if self._conn is None:
-            self._conn = sqlite.connect(self._db_file)
+            if not self.conf.readable:
+                return None
+
+            try:
+                self._conn = sqlite.connect(self._db_file)
+            except (sqlite.OperationalError, sqlite.DatabaseError):
+                self.conf.readable = False
+                return None
+
         return self._conn.cursor()
     def _commit(self):
         return self._conn.commit()
@@ -286,7 +344,7 @@ class YumHistory:
     def _ipkg2pid(self, po):
         csum = None
         yumdb = po.yumdb_info
-        if 'checksum_type' in yumdb and 'checksum_type' in yumdb:
+        if 'checksum_type' in yumdb and 'checksum_data' in yumdb:
             csum = "%s:%s" % (yumdb.checksum_type, yumdb.checksum_data)
         return self._pkgtup2pid(po.pkgtup, csum)
     def pkg2pid(self, po):
@@ -300,7 +358,7 @@ class YumHistory:
     def txmbr2state(txmbr):
         state = None
         if txmbr.output_state in (TS_INSTALL, TS_TRUEINSTALL):
-            if hasattr(txmbr, 'reinstall'):
+            if txmbr.reinstall:
                 state = 'Reinstall'
             elif txmbr.downgrades:
                 state = 'Downgrade'
@@ -315,17 +373,32 @@ class YumHistory:
 
     def trans_with_pid(self, pid):
         cur = self._get_cursor()
+        if cur is None:
+            return None
         res = executeSQL(cur,
                          """INSERT INTO trans_with_pkgs
                          (tid, pkgtupid)
                          VALUES (?, ?)""", (self._tid, pid))
         return cur.lastrowid
 
+    def trans_skip_pid(self, pid):
+        cur = self._get_cursor()
+        if cur is None or not self._update_db_file_2():
+            return None
+        
+        res = executeSQL(cur,
+                         """INSERT INTO trans_skip_pkgs
+                         (tid, pkgtupid)
+                         VALUES (?, ?)""", (self._tid, pid))
+        return cur.lastrowid
+
     def trans_data_pid_beg(self, pid, state):
         assert state is not None
         if not hasattr(self, '_tid') or state is None:
             return # Not configured to run
         cur = self._get_cursor()
+        if cur is None:
+            return # Should never happen, due to above
         res = executeSQL(cur,
                          """INSERT INTO trans_data_pkgs
                          (tid, pkgtupid, state)
@@ -337,6 +410,8 @@ class YumHistory:
             return # Not configured to run
 
         cur = self._get_cursor()
+        if cur is None:
+            return # Should never happen, due to above
         res = executeSQL(cur,
                          """UPDATE trans_data_pkgs SET done = ?
                          WHERE tid = ? AND pkgtupid = ? AND state = ?
@@ -344,8 +419,62 @@ class YumHistory:
         self._commit()
         return cur.lastrowid
 
-    def beg(self, rpmdb_version, using_pkgs, txmbrs):
+    def _trans_rpmdb_problem(self, problem):
+        if not hasattr(self, '_tid'):
+            return # Not configured to run
         cur = self._get_cursor()
+        if cur is None or not self._update_db_file_2():
+            return None
+        res = executeSQL(cur,
+                         """INSERT INTO trans_rpmdb_problems
+                         (tid, problem, msg)
+                         VALUES (?, ?, ?)""", (self._tid,
+                                               problem.problem,
+                                               to_unicode(str(problem))))
+        rpid = cur.lastrowid
+
+        if not rpid:
+            return rpid
+
+        pkgs = {}
+        pkg = problem.pkg
+        pkgs[pkg.pkgtup] = pkg
+        if problem.problem == 'conflicts':
+            for pkg in problem.conflicts:
+                pkgs[pkg.pkgtup] = pkg
+        if problem.problem == 'duplicates':
+            pkgs[problem.duplicate.pkgtup] = problem.duplicate
+
+        for pkg in pkgs.values():
+            pid = self.pkg2pid(pkg)
+            if pkg.pkgtup == problem.pkg.pkgtup:
+                main = 'TRUE'
+            else:
+                main = 'FALSE'
+            res = executeSQL(cur,
+                             """INSERT INTO trans_prob_pkgs
+                             (rpid, pkgtupid, main)
+                             VALUES (?, ?, ?)""", (rpid, pid, main))
+
+        return rpid
+
+    def _trans_cmdline(self, cmdline):
+        if not hasattr(self, '_tid'):
+            return # Not configured to run
+        cur = self._get_cursor()
+        if cur is None or not self._update_db_file_2():
+            return None
+        res = executeSQL(cur,
+                         """INSERT INTO trans_cmdline
+                         (tid, cmdline)
+                         VALUES (?, ?)""", (self._tid, cmdline))
+        return cur.lastrowid
+
+    def beg(self, rpmdb_version, using_pkgs, txmbrs, skip_packages=[],
+            rpmdb_problems=[], cmdline=None):
+        cur = self._get_cursor()
+        if cur is None:
+            return
         res = executeSQL(cur,
                          """INSERT INTO trans_beg
                             (timestamp, rpmdb_version, loginuid)
@@ -363,10 +492,22 @@ class YumHistory:
             state = self.txmbr2state(txmbr)
             self.trans_data_pid_beg(pid, state)
         
+        for pkg in skip_packages:
+            pid   = self.pkg2pid(pkg)
+            self.trans_skip_pid(pid)
+
+        for problem in rpmdb_problems:
+            self._trans_rpmdb_problem(problem)
+
+        if cmdline:
+            self._trans_cmdline(cmdline)
+
         self._commit()
 
     def _log_errors(self, errors):
         cur = self._get_cursor()
+        if cur is None:
+            return
         for error in errors:
             error = to_unicode(error)
             executeSQL(cur,
@@ -380,7 +521,9 @@ class YumHistory:
             return # Not configured to run
 
         cur = self._get_cursor()
-        for error in msg.split('\n'):
+        if cur is None:
+            return # Should never happen, due to above
+        for error in msg.splitlines():
             error = to_unicode(error)
             executeSQL(cur,
                        """INSERT INTO trans_script_stdout
@@ -411,7 +554,11 @@ class YumHistory:
 
     def end(self, rpmdb_version, return_code, errors=None):
         assert return_code or not errors
+        if not hasattr(self, '_tid'):
+            return # Failed at beg() time
         cur = self._get_cursor()
+        if cur is None:
+            return # Should never happen, due to above
         res = executeSQL(cur,
                          """INSERT INTO trans_end
                             (tid, timestamp, rpmdb_version, return_code)
@@ -431,6 +578,63 @@ class YumHistory:
             self._log_errors(errors)
         del self._tid
 
+    def write_addon_data(self, dataname, data):
+        """append data to an arbitrary-named file in the history 
+           addon_path/transaction id location,
+           returns True if write succeeded, False if not"""
+        
+        if not hasattr(self, '_tid'):
+            # maybe we should raise an exception or a warning here?
+            return False
+        
+        if not dataname:
+            return False
+        
+        if not data:
+            return False
+            
+        # make sure the tid dir exists
+        tid_dir = self.conf.addon_path + '/' + str(self._tid)
+
+        if self.conf.writable and not os.path.exists(tid_dir):
+            try:
+                os.makedirs(tid_dir, mode=0700)
+            except (IOError, OSError), e:
+                # emit a warning/raise an exception?
+                return False
+        
+        # cleanup dataname
+        safename = dataname.replace('/', '_')
+        data_fn = tid_dir + '/' + safename
+        try:
+            # open file in append
+            fo = open(data_fn, 'w+')
+            # write data
+            fo.write(data)
+            # flush data
+            fo.flush()
+            fo.close()
+        except (IOError, OSError), e:
+            return False
+        # return
+        return True
+        
+    def return_addon_data(self, tid, item=None):
+        hist_and_tid = self.conf.addon_path + '/' + str(tid) + '/'
+        addon_info = glob.glob(hist_and_tid + '*')
+        addon_names = [ i.replace(hist_and_tid, '') for i in addon_info ]
+        if not item:
+            return addon_names
+        
+        if item not in addon_names:
+            # XXX history needs SOME kind of exception, or warning, I think?
+            return None
+        
+        fo = open(hist_and_tid + item, 'r')
+        data = fo.read()
+        fo.close()
+        return data
+        
     def _old_with_pkgs(self, tid):
         cur = self._get_cursor()
         executeSQL(cur,
@@ -463,11 +667,70 @@ class YumHistory:
                 obj.state_installed = False
             ret.append(obj)
         return ret
+    def _old_skip_pkgs(self, tid):
+        cur = self._get_cursor()
+        if cur is None or not self._update_db_file_2():
+            return []
+        executeSQL(cur,
+                   """SELECT name, arch, epoch, version, release, checksum
+                      FROM trans_skip_pkgs JOIN pkgtups USING(pkgtupid)
+                      WHERE tid = ?
+                      ORDER BY name ASC, epoch ASC""", (tid,))
+        ret = []
+        for row in cur:
+            obj = YumHistoryPackage(row[0],row[1],row[2],row[3],row[4], row[5])
+            ret.append(obj)
+        return ret
+    def _old_prob_pkgs(self, rpid):
+        cur = self._get_cursor()
+        if cur is None or not self._update_db_file_2():
+            return []
+        executeSQL(cur,
+                   """SELECT name, arch, epoch, version, release, checksum, main
+                      FROM trans_prob_pkgs JOIN pkgtups USING(pkgtupid)
+                      WHERE rpid = ?
+                      ORDER BY name ASC, epoch ASC""", (rpid,))
+        ret = []
+        for row in cur:
+            obj = YumHistoryPackage(row[0],row[1],row[2],row[3],row[4], row[5])
+            obj.main = row[6] == 'TRUE'
+            ret.append(obj)
+        return ret
+
+    def _old_problems(self, tid):
+        cur = self._get_cursor()
+        if cur is None or not self._update_db_file_2():
+            return []
+        executeSQL(cur,
+                   """SELECT rpid, problem, msg
+                      FROM trans_rpmdb_problems
+                      WHERE tid = ?
+                      ORDER BY problem ASC, rpid ASC""", (tid,))
+        ret = []
+        for row in cur:
+            obj = YumHistoryRpmdbProblem(self, row[0], row[1], row[2])
+            ret.append(obj)
+        return ret
+
+    def _old_cmdline(self, tid):
+        cur = self._get_cursor()
+        if cur is None or not self._update_db_file_2():
+            return None
+        executeSQL(cur,
+                   """SELECT cmdline
+                      FROM trans_cmdline
+                      WHERE tid = ?""", (tid,))
+        ret = []
+        for row in cur:
+            return row[0]
+        return None
 
     def old(self, tids=[], limit=None, complete_transactions_only=False):
         """ Return a list of the last transactions, note that this includes
             partial transactions (ones without an end transaction). """
         cur = self._get_cursor()
+        if cur is None:
+            return []
         sql =  """SELECT tid,
                          trans_beg.timestamp AS beg_ts,
                          trans_beg.rpmdb_version AS beg_rv,
@@ -539,23 +802,14 @@ class YumHistory:
 
         return ret
 
-    def last(self):
-        """ This is the last full transaction. So any imcomplete transactions
-            do not count. """
-        cur = self._get_cursor()
-        sql =  """SELECT tid,
-                         trans_beg.timestamp AS beg_ts,
-                         trans_beg.rpmdb_version AS beg_rv,
-                         trans_end.timestamp AS end_ts,
-                         trans_end.rpmdb_version AS end_rv,
-                         loginuid, return_code
-                  FROM trans_beg JOIN trans_end USING(tid)
-                  ORDER BY beg_ts DESC, tid ASC
-                  LIMIT 1"""
-        executeSQL(cur, sql)
-        for row in cur:
-            return YumHistoryTransaction(self, row)
-        return None
+    def last(self, complete_transactions_only=True):
+        """ This is the last full transaction. So any incomplete transactions
+            do not count, by default. """
+        ret = self.old([], 1, complete_transactions_only)
+        if not ret:
+            return None
+        assert len(ret) == 1
+        return ret[0]
 
     def _yieldSQLDataList(self, patterns, fields, ignore_case):
         """Yields all the package data for the given params. """
@@ -584,6 +838,10 @@ class YumHistory:
             packages al. la. "yum list". Returns transaction ids. """
         # Search packages ... kind of sucks that it's search not list, pkglist?
 
+        cur = self._get_cursor()
+        if cur is None:
+            return set()
+
         data = _setupHistorySearchSQL(patterns, ignore_case)
         (need_full, patterns, fields, names) = data
 
@@ -592,7 +850,6 @@ class YumHistory:
         for row in self._yieldSQLDataList(patterns, fields, ignore_case):
             pkgtupids.add(row[0])
 
-        cur = self._get_cursor()
         sql =  """SELECT tid FROM trans_data_pkgs WHERE pkgtupid IN """
         sql += "(%s)" % ",".join(['?'] * len(pkgtupids))
         params = list(pkgtupids)
@@ -610,16 +867,108 @@ class YumHistory:
             tids.add(row[0])
         return tids
 
+    _update_ops_2 = ['''\
+\
+ CREATE TABLE trans_skip_pkgs (
+     tid INTEGER NOT NULL REFERENCES trans_beg,
+     pkgtupid INTEGER NOT NULL REFERENCES pkgtups);
+''', '''\
+\
+ CREATE TABLE trans_cmdline (
+     tid INTEGER NOT NULL REFERENCES trans_beg,
+     cmdline TEXT NOT NULL);
+''', '''\
+\
+ CREATE TABLE trans_rpmdb_problems (
+     rpid INTEGER PRIMARY KEY,
+     tid INTEGER NOT NULL REFERENCES trans_beg,
+     problem TEXT NOT NULL, msg TEXT NOT NULL);
+''', '''\
+\
+ CREATE TABLE trans_prob_pkgs (
+     rpid INTEGER NOT NULL REFERENCES trans_rpmdb_problems,
+     pkgtupid INTEGER NOT NULL REFERENCES pkgtups,
+     main BOOL NOT NULL DEFAULT FALSE);
+''', '''\
+\
+ CREATE VIEW vtrans_data_pkgs AS
+     SELECT tid,name,epoch,version,release,arch,pkgtupid,
+            state,done,
+            name || '-' || epoch || ':' ||
+            version || '-' || release || '.' || arch AS nevra
+     FROM trans_data_pkgs JOIN pkgtups USING(pkgtupid)
+     ORDER BY name;
+''', '''\
+\
+ CREATE VIEW vtrans_with_pkgs AS
+     SELECT tid,name,epoch,version,release,arch,pkgtupid,
+            name || '-' || epoch || ':' ||
+            version || '-' || release || '.' || arch AS nevra
+     FROM trans_with_pkgs JOIN pkgtups USING(pkgtupid)
+     ORDER BY name;
+''', '''\
+\
+ CREATE VIEW vtrans_skip_pkgs AS
+     SELECT tid,name,epoch,version,release,arch,pkgtupid,
+            name || '-' || epoch || ':' ||
+            version || '-' || release || '.' || arch AS nevra
+     FROM trans_skip_pkgs JOIN pkgtups USING(pkgtupid)
+     ORDER BY name;
+''', # NOTE: Old versions of sqlite don't like this next view, they are only
+     #       there for debugging anyway ... but it's worth remembering.
+'''\
+\
+ CREATE VIEW vtrans_prob_pkgs AS
+     SELECT tid,rpid,name,epoch,version,release,arch,pkgtups.pkgtupid,
+            main,
+            name || '-' || epoch || ':' ||
+            version || '-' || release || '.' || arch AS nevra
+     FROM (trans_prob_pkgs JOIN trans_rpmdb_problems USING(rpid))
+                           JOIN pkgtups USING(pkgtupid)
+     ORDER BY name;
+''']
+
+    def _update_db_file_2(self):
+        """ Update to version 2 of history, includes trans_skip_pkgs. """
+        if not self.conf.writable:
+            return False
+
+        if hasattr(self, '_cached_updated_2'):
+            return self._cached_updated_2
+
+        cur = self._get_cursor()
+        if cur is None:
+            return False
+
+        executeSQL(cur, "PRAGMA table_info(trans_skip_pkgs)")
+        #  If we get anything, we're fine. There might be a better way of
+        # saying "anything" but this works.
+        for ob in cur:
+            break
+        else:
+            for op in self._update_ops_2:
+                cur.execute(op)
+            self._commit()
+        self._cached_updated_2 = True
+        return True
+
     def _create_db_file(self):
         """ Create a new history DB file, populating tables etc. """
 
+        self._db_date = time.strftime('%Y-%m-%d')
         _db_file = '%s/%s-%s.%s' % (self.conf.db_path,
                                     'history',
-                                    time.strftime('%Y-%m-%d'),
+                                    self._db_date,
                                     'sqlite')
         if self._db_file == _db_file:
             os.rename(_db_file, _db_file + '.old')
         self._db_file = _db_file
+        
+        if self.conf.writable and not os.path.exists(self._db_file):
+            # make them default to 0600 - sysadmin can change it later
+            # if they want
+            fo = os.open(self._db_file, os.O_CREAT, 0600)
+            os.close(fo)
                 
         cur = self._get_cursor()
         ops = ['''\
@@ -665,6 +1014,8 @@ class YumHistory:
 ''']
         for op in ops:
             cur.execute(op)
+        for op in self._update_ops_2:
+            cur.execute(op)
         self._commit()
 
 # Pasted from sqlitesack
