@@ -23,6 +23,7 @@ import os
 import cli
 from yum import logginglevels
 from yum import _
+from yum import misc
 import yum.Errors
 import operator
 import locale
@@ -84,7 +85,7 @@ def checkGroupArg(base, basecmd, extcmds):
 
 def checkCleanArg(base, basecmd, extcmds):
     VALID_ARGS = ('headers', 'packages', 'metadata', 'dbcache', 'plugins',
-                  'expire-cache', 'all')
+                  'expire-cache', 'rpmdb', 'all')
 
     if len(extcmds) == 0:
         base.logger.critical(_('Error: clean requires an option: %s') % (
@@ -204,15 +205,34 @@ class UpdateCommand(YumCommand):
         except yum.Errors.YumBaseError, e:
             return 1, [str(e)]
 
+class DistroSyncCommand(YumCommand):
+    def getNames(self):
+        return ['distribution-synchronization', 'distro-sync']
+
+    def getUsage(self):
+        return _("[PACKAGE...]")
+
+    def getSummary(self):
+        return _("Synchronize installed packages to the latest available versions")
+
+    def doCheck(self, base, basecmd, extcmds):
+        checkRootUID(base)
+        checkGPGKey(base)
+
+    def doCommand(self, base, basecmd, extcmds):
+        self.doneCommand(base, _("Setting up Distribution Synchronization Process"))
+        try:
+            base.conf.obsoletes = 1
+            return base.distroSyncPkgs(extcmds)
+        except yum.Errors.YumBaseError, e:
+            return 1, [str(e)]
+
 def _add_pkg_simple_list_lens(data, pkg, indent=''):
     """ Get the length of each pkg's column. Add that to data.
         This "knows" about simpleList and printVer. """
     na  = len(pkg.name)    + 1 + len(pkg.arch)    + len(indent)
     ver = len(pkg.version) + 1 + len(pkg.release)
-    if pkg.repoid == 'installed' and 'from_repo' in pkg.yumdb_info:
-        rid = len(pkg.yumdb_info.from_repo) + 1
-    else:
-        rid = len(pkg.repoid)
+    rid = len(pkg.ui_from_repo)
     if pkg.epoch != '0':
         ver += len(pkg.epoch) + 1
     for (d, v) in (('na', na), ('ver', ver), ('rid', rid)):
@@ -774,6 +794,9 @@ class RepoListCommand(YumCommand):
                     return True
             return False
 
+        def _num2ui_num(num):
+            return to_unicode(locale.format("%d", num, True))
+
         if len(extcmds) >= 1 and extcmds[0] in ('all', 'disabled', 'enabled'):
             arg = extcmds[0]
             extcmds = extcmds[1:]
@@ -782,49 +805,80 @@ class RepoListCommand(YumCommand):
         extcmds = map(lambda x: x.lower(), extcmds)
 
         verbose = base.verbose_logger.isEnabledFor(logginglevels.DEBUG_3)
-        try:
-            # Setup so len(repo.sack) is correct
-            base.repos.populateSack()
-        except yum.Errors.RepoError:
-            if verbose:
-                raise
+        if arg != 'disabled' or extcmds:
+            try:
+                # Setup so len(repo.sack) is correct
+                base.repos.populateSack()
+                base.pkgSack # Need to setup the pkgSack, so excludes work
+            except yum.Errors.RepoError:
+                if verbose:
+                    raise
 
         repos = base.repos.repos.values()
         repos.sort()
         enabled_repos = base.repos.listEnabled()
-        if arg == 'all':
-            ehibeg = base.term.FG_COLOR['green'] + base.term.MODE['bold']
-            dhibeg = base.term.FG_COLOR['red']
-            hiend  = base.term.MODE['normal']
-        else:
-            ehibeg = ''
-            dhibeg = ''
-            hiend  = ''
+        on_ehibeg = base.term.FG_COLOR['green'] + base.term.MODE['bold']
+        on_dhibeg = base.term.FG_COLOR['red']
+        on_hiend  = base.term.MODE['normal']
         tot_num = 0
         cols = []
         for repo in repos:
             if len(extcmds) and not _repo_match(repo, extcmds):
                 continue
+            (ehibeg, dhibeg, hiend)  = '', '', ''
+            ui_enabled      = ''
+            ui_endis_wid    = 0
+            ui_num          = ""
+            ui_excludes_num = ''
+            force_show = False
+            if arg == 'all' or repo.id in extcmds or repo.name in extcmds:
+                force_show = True
+                (ehibeg, dhibeg, hiend) = (on_ehibeg, on_dhibeg, on_hiend)
             if repo in enabled_repos:
                 enabled = True
-                ui_enabled = ehibeg + _('enabled') + hiend + ": "
-                ui_endis_wid = utf8_width(_('enabled')) + 2
-                num        = len(repo.sack)
-                tot_num   += num
-                ui_num     = to_unicode(locale.format("%d", num, True))
+                if arg == 'enabled':
+                    force_show = False
+                elif arg == 'disabled' and not force_show:
+                    continue
+                if force_show or verbose:
+                    ui_enabled = ehibeg + _('enabled') + hiend
+                    ui_endis_wid = utf8_width(_('enabled'))
+                    if not verbose:
+                        ui_enabled += ": "
+                        ui_endis_wid += 2
                 if verbose:
                     ui_size = _repo_size(repo)
+                # We don't show status for list disabled
+                if arg != 'disabled' or verbose:
+                    if verbose or base.conf.exclude or repo.exclude:
+                        num        = len(repo.sack.simplePkgList())
+                    else:
+                        num        = len(repo.sack)
+                    ui_num     = _num2ui_num(num)
+                    excludes   = repo.sack._excludes
+                    excludes   = len([pid for r,pid in excludes if r == repo])
+                    if excludes:
+                        ui_excludes_num = _num2ui_num(excludes)
+                        if not verbose:
+                            ui_num += "+%s" % ui_excludes_num
+                    tot_num   += num
             else:
                 enabled = False
+                if arg == 'disabled':
+                    force_show = False
+                elif arg == 'enabled' and not force_show:
+                    continue
                 ui_enabled = dhibeg + _('disabled') + hiend
                 ui_endis_wid = utf8_width(_('disabled'))
-                ui_num     = ""
-                
-            if (arg == 'all' or
-                (arg == 'enabled' and enabled) or
-                (arg == 'disabled' and not enabled)):
+
+            if True: # Here to make patch smaller, TODO: rm
                 if not verbose:
-                    cols.append((str(repo), repo.name,
+                    rid = str(repo)
+                    if enabled and repo.metalink:
+                        mdts = repo.metalink_data.repomd.timestamp
+                        if mdts > repo.repoXML.timestamp:
+                            rid = '*' + rid
+                    cols.append((rid, repo.name,
                                  (ui_enabled, ui_endis_wid), ui_num))
                 else:
                     if enabled:
@@ -832,8 +886,11 @@ class RepoListCommand(YumCommand):
                     else:
                         md = None
                     out = [base.fmtKeyValFill(_("Repo-id      : "), repo),
-                           base.fmtKeyValFill(_("Repo-name    : "), repo.name),
-                           base.fmtKeyValFill(_("Repo-status  : "), ui_enabled)]
+                           base.fmtKeyValFill(_("Repo-name    : "), repo.name)]
+
+                    if force_show or extcmds:
+                        out += [base.fmtKeyValFill(_("Repo-status  : "),
+                                                   ui_enabled)]
                     if md and md.revision is not None:
                         out += [base.fmtKeyValFill(_("Repo-revision: "),
                                                    md.revision)]
@@ -889,7 +946,7 @@ class RepoListCommand(YumCommand):
                     elif not repo.metadata_expire:
                         num = _("Instant (last: %s)") % last
                     else:
-                        num = locale.format("%d", repo.metadata_expire, True)
+                        num = _num2ui_num(repo.metadata_expire)
                         num = _("%s second(s) (last: %s)") % (num, last)
 
                     out += [base.fmtKeyValFill(_("Repo-expire  : "), num)]
@@ -902,16 +959,20 @@ class RepoListCommand(YumCommand):
                         out += [base.fmtKeyValFill(_("Repo-include : "),
                                                    ", ".join(repo.includepkgs))]
 
+                    if ui_excludes_num:
+                        out += [base.fmtKeyValFill(_("Repo-excluded: "),
+                                                   ui_excludes_num)]
+
                     base.verbose_logger.log(logginglevels.DEBUG_3,
                                             "%s\n",
-                                            "\n".join(out))
+                                            "\n".join(map(misc.to_unicode, out)))
 
         if not verbose and cols:
             #  Work out the first (id) and last (enabled/disalbed/count),
             # then chop the middle (name)...
             id_len = utf8_width(_('repo id'))
             nm_len = 0
-            ct_len = 0
+            st_len = 0
             ui_len = 0
 
             for (rid, rname, (ui_enabled, ui_endis_wid), ui_num) in cols:
@@ -919,14 +980,17 @@ class RepoListCommand(YumCommand):
                     id_len = utf8_width(rid)
                 if nm_len < utf8_width(rname):
                     nm_len = utf8_width(rname)
-                if ct_len < ui_endis_wid:
-                    ct_len = ui_endis_wid
+                if st_len < (ui_endis_wid + len(ui_num)):
+                    st_len = (ui_endis_wid + len(ui_num))
+                # Need this as well as above for: utf8_width_fill()
                 if ui_len < len(ui_num):
                     ui_len = len(ui_num)
-            if utf8_width(_('status')) > ct_len + ui_len:
+            if arg == 'disabled': # Don't output a status column.
+                left = base.term.columns - (id_len + 1)
+            elif utf8_width(_('status')) > st_len:
                 left = base.term.columns - (id_len + utf8_width(_('status')) +2)
             else:
-                left = base.term.columns - (id_len + ct_len + ui_len + 2)
+                left = base.term.columns - (id_len + st_len + 2)
 
             if left < nm_len: # Name gets chopped
                 nm_len = left
@@ -937,9 +1001,20 @@ class RepoListCommand(YumCommand):
 
             txt_rid  = utf8_width_fill(_('repo id'), id_len)
             txt_rnam = utf8_width_fill(_('repo name'), nm_len, nm_len)
-            base.verbose_logger.log(logginglevels.INFO_2,"%s %s %s",
-                                    txt_rid, txt_rnam, _('status'))
+            if arg == 'disabled': # Don't output a status column.
+                base.verbose_logger.log(logginglevels.INFO_2,"%s %s",
+                                        txt_rid, txt_rnam)
+            else:
+                base.verbose_logger.log(logginglevels.INFO_2,"%s %s %s",
+                                        txt_rid, txt_rnam, _('status'))
             for (rid, rname, (ui_enabled, ui_endis_wid), ui_num) in cols:
+                if arg == 'disabled': # Don't output a status column.
+                    base.verbose_logger.log(logginglevels.INFO_2, "%s %s",
+                                            utf8_width_fill(rid, id_len),
+                                            utf8_width_fill(rname, nm_len,
+                                                            nm_len))
+                    continue
+
                 if ui_num:
                     ui_num = utf8_width_fill(ui_num, ui_len, left=False)
                 base.verbose_logger.log(logginglevels.INFO_2, "%s %s %s%s",
@@ -1009,7 +1084,7 @@ class HelpCommand(YumCommand):
         return help_output
 
     def doCommand(self, base, basecmd, extcmds):
-        if base.yum_cli_commands.has_key(extcmds[0]):
+        if extcmds[0] in base.yum_cli_commands:
             command = base.yum_cli_commands[extcmds[0]]
             base.verbose_logger.log(logginglevels.INFO_2,
                     self._makeOutput(command))
@@ -1101,20 +1176,29 @@ class VersionCommand(YumCommand):
 
         verbose = base.verbose_logger.isEnabledFor(logginglevels.DEBUG_3)
         groups = {}
-        gconf = yum.config.readVersionGroupsConfig()
+        if vcmd in ('nogroups', 'nogroups-installed', 'nogroups-available',
+                    'nogroups-all'):
+            gconf = []
+            if vcmd == 'nogroups':
+                vcmd = 'installed'
+            else:
+                vcmd = vcmd[len('nogroups-'):]
+        else:
+            gconf = yum.config.readVersionGroupsConfig()
+
         for group in gconf:
             groups[group] = set(gconf[group].pkglist)
             if gconf[group].run_with_packages:
                 groups[group].update(base.run_with_package_names)
 
-        if vcmd in ('grouplist'):
+        if vcmd == 'grouplist':
             print _(" Yum version groups:")
             for group in sorted(groups):
                 print "   ", group
 
             return 0, ['version grouplist']
 
-        if vcmd in ('groupinfo'):
+        if vcmd == 'groupinfo':
             for group in groups:
                 if group not in extcmds[1:]:
                     continue
@@ -1136,8 +1220,8 @@ class VersionCommand(YumCommand):
 
             return 0, ['version groupinfo']
 
-        rel = base.yumvar['releasever']
-        ba  = base.yumvar['basearch']
+        rel = base.conf.yumvar['releasever']
+        ba  = base.conf.yumvar['basearch']
         cols = []
         if vcmd in ('installed', 'all', 'group-installed', 'group-all'):
             try:
@@ -1146,14 +1230,16 @@ class VersionCommand(YumCommand):
                 if lastdbv is not None:
                     lastdbv = lastdbv.end_rpmdbversion
                 if lastdbv is not None and data[0] != lastdbv:
-                    errstring = _('Warning: RPMDB has been altered since the last yum transaction.')
-                    base.logger.warning(errstring)
+                    base._rpmdb_warn_checks(warn=lastdbv is not None)
                 if vcmd not in ('group-installed', 'group-all'):
                     cols.append(("%s %s/%s" % (_("Installed:"), rel, ba),
                                  str(data[0])))
                     _append_repos(cols, data[1])
                 if groups:
                     for grp in sorted(data[2]):
+                        if (vcmd.startswith("group-") and
+                            len(extcmds) > 1 and grp not in extcmds[1:]):
+                            continue
                         cols.append(("%s %s" % (_("Group-Installed:"), grp),
                                      str(data[2][grp])))
                         _append_repos(cols, data[3][grp])
@@ -1169,6 +1255,9 @@ class VersionCommand(YumCommand):
                         _append_repos(cols, data[1])
                 if groups:
                     for grp in sorted(data[2]):
+                        if (vcmd.startswith("group-") and
+                            len(extcmds) > 1 and grp not in extcmds[1:]):
+                            continue
                         cols.append(("%s %s" % (_("Group-Available:"), grp),
                                      str(data[2][grp])))
                         if verbose:
@@ -1234,7 +1323,7 @@ class HistoryCommand(YumCommand):
         base.history._create_db_file()
 
     def doCheck(self, base, basecmd, extcmds):
-        cmds = ('list', 'info', 'summary', 'repeat', 'redo', 'undo', 'new')
+        cmds = ('list', 'info', 'summary', 'repeat', 'redo', 'undo', 'new', 'addon-info')
         if extcmds and extcmds[0] not in cmds:
             base.logger.critical(_('Invalid history sub-command, use: %s.'),
                                  ", ".join(cmds))
@@ -1242,6 +1331,9 @@ class HistoryCommand(YumCommand):
         if extcmds and extcmds[0] in ('repeat', 'redo', 'undo', 'new'):
             checkRootUID(base)
             checkGPGKey(base)
+        elif not os.access(base.history._db_file, os.R_OK):
+            base.logger.critical(_("You don't have access to the history DB."))
+            raise cli.CliError
 
     def doCommand(self, base, basecmd, extcmds):
         vcmd = 'list'
@@ -1255,6 +1347,8 @@ class HistoryCommand(YumCommand):
             ret = base.historyInfoCmd(extcmds)
         elif vcmd == 'summary':
             ret = base.historySummaryCmd(extcmds)
+        elif vcmd == 'addon-info':
+            ret = base.historyAddonInfoCmd(extcmds)
         elif vcmd == 'undo':
             ret = self._hcmd_undo(base, extcmds)
         elif vcmd in ('redo', 'repeat'):
@@ -1271,3 +1365,32 @@ class HistoryCommand(YumCommand):
         if extcmds:
             vcmd = extcmds[0]
         return vcmd in ('repeat', 'redo', 'undo')
+
+
+class CheckRpmdbCommand(YumCommand):
+    def getNames(self):
+        return ['check', 'check-rpmdb']
+
+    def getUsage(self):
+        return "[dependencies|duplicates|all]"
+
+    def getSummary(self):
+        return _("Check for problems in the rpmdb")
+
+    def doCommand(self, base, basecmd, extcmds):
+        chkcmd = 'all'
+        if extcmds:
+            chkcmd = extcmds
+
+        def _out(x):
+            print x
+
+        rc = 0
+        if base._rpmdb_warn_checks(out=_out, warn=False, chkcmd=chkcmd,
+                                   header=lambda x: None):
+            rc = 1
+        return rc, ['%s %s' % (basecmd, chkcmd)]
+
+    def needTs(self, base, basecmd, extcmds):
+        return False
+
